"""
### Building awesome visualizations for algorithms

Although on paper one algorithm might prove better than other but it's mostly based on nature of order of increase in 
running time with respect to input size.

However, in practice an algorithm having higher runtime complexity than others may actually have a smaller runtime
for your specific test case.

With gopy, you can test each algorithm's behavior for your specific input and test case and compare actual running times in practice.

eg:

# test for knuth_morris_pratt

```python
from gopy.profile import profile
from gopy.strings.knuth_morris_pratt import match 
print(profile('match("ABCDAADDABCABAB","A")'))
```

This will make in depth visualizations in your browser for the kmp algorithm.
"""

import cProfile
import subprocess

def profile(s):
    cProfile.run(s,'test.profile')
    return subprocess.check_output(['snakeviz', 'test.profile'])


"""
Jump search technique also works for ordered lists. It creates a block and 
tries to find the element in that block. If the item is not in the block, 
it shifts the entire block. The block size is based on the size of the list. 
If the size of the list is n then block size will be √n. 

After finding a correct block it finds the item using a linear search technique. The jump 
search lies between linear search and binary search according to its performance.

### The complexity of Jump Search Technique

- Time Complexity: O(√n)

- Space Complexity: O(1)

### Input and Output

```
Input: 

A sorted list of data:
10 13 15 26 28 50 56 88 94 127 159 356 480 567 689 699 780 850 956 995
The search key 356

Output: 

Item found at location: 11
```

### Algorithm

```
jumpSearch(array, size, key)

Input: An sorted array, size of the array and the search key

Output: location of the key (if found), otherwise wrong location.

Begin
   blockSize := √size
   start := 0
   end := blockSize
   while array[end] <= key AND end < size do
      start := end
      end := end + blockSize
      if end > size – 1 then
         end := size
   done
   for i := start to end -1 do
      if array[i] = key then
         return i
   done
   return invalid location
End
```
"""

import math 
  
def search( x , arr ): 
    n = len(arr)
    # Finding block size to be jumped 
    step = int(math.sqrt(n)) 
      
    # Finding the block where element is 
    # present (if it is present) 
    prev = 0
    while arr[int(min(step, n)-1)] < x: 
        prev = step 
        step += int(math.sqrt(n)) 
        if prev >= n: 
            return "Not Found"
      
    # Doing a linear search for x in  
    # block beginning with prev. 
    while arr[int(prev)] < x: 
        prev += 1
          
        # If we reached next block or end  
        # of array, element is not present. 
        if prev == min(step, n): 
            return "Not Found"
      
    # If element is found 
    if arr[int(prev)] == x: 
        return prev 
      
    return "Not Found"


# Given an m x n matrix of non-negative integers representing
# the height of each unit cell in a continent,
# the "Pacific ocean" touches the left and top edges of the matrix
# and the "Atlantic ocean" touches the right and bottom edges.

# Water can only flow in four directions (up, down, left, or right)
# from a cell to another one with height equal or lower.

# Find the list of grid coordinates where water can flow to both the
# Pacific and Atlantic ocean.

# Note:
# The order of returned grid coordinates does not matter.
# Both m and n are less than 150.
# Example:

# Given the following 5x5 matrix:

  # Pacific ~   ~   ~   ~   ~
       # ~  1   2   2   3  (5) *
       # ~  3   2   3  (4) (4) *
       # ~  2   4  (5)  3   1  *
       # ~ (6) (7)  1   4   5  *
       # ~ (5)  1   1   2   4  *
          # *   *   *   *   * Atlantic

# Return:

# [[0, 4], [1, 3], [1, 4], [2, 2], [3, 0], [3, 1], [4, 0]]
# (positions with parentheses in above matrix).

def pacific_atlantic(matrix):
    """
    :type matrix: List[List[int]]
    :rtype: List[List[int]]
    """
    n = len(matrix)
    if not n: return []
    m = len(matrix[0])
    if not m: return []
    res = []
    atlantic = [[False for _ in range (n)] for _ in range(m)]
    pacific =  [[False for _ in range (n)] for _ in range(m)]
    for i in range(n):
        dfs(pacific, matrix, float("-inf"), i, 0)
        dfs(atlantic, matrix, float("-inf"), i, m-1)
    for i in range(m):
        dfs(pacific, matrix, float("-inf"), 0, i)
        dfs(atlantic, matrix, float("-inf"), n-1, i)
    for i in range(n):
        for j in range(m):
            if pacific[i][j] and atlantic[i][j]:
                res.append([i, j])
    return res

def dfs(grid, matrix, height, i, j):
    if i < 0 or i >= len(matrix) or j < 0 or  j >= len(matrix[0]):
        return
    if grid[i][j] or matrix[i][j] < height:
        return
    grid[i][j] = True
    dfs(grid, matrix, matrix[i][j], i-1, j)
    dfs(grid, matrix, matrix[i][j], i+1, j)
    dfs(grid, matrix, matrix[i][j], i, j-1)
    dfs(grid, matrix, matrix[i][j], i, j+1)



"""
It's similar to how human solve Sudoku.

create a hash table (dictionary) val to store possible values in every location.
Each time, start from the location with fewest possible values, choose one value
from it and then update the board and possible values at other locations.
If this update is valid, keep solving (DFS). If this update is invalid (leaving
zero possible values at some locations) or this value doesn't lead to the
solution, undo the updates and then choose the next value.
Since we calculated val at the beginning and start filling the board from the
location with fewest possible values, the amount of calculation and thus the
runtime can be significantly reduced:


The run time is 48-68 ms on LeetCode OJ, which seems to be among the fastest
python solutions here.


The PossibleVals function may be further simplified/optimized, but it works just
fine for now. (it would look less lengthy if we are allowed to use numpy array
for the board lol).
"""
class Sudoku: 
    def __init__ (self, board, row, col):
        self.board = board
        self.row = row
        self.col = col
        self.val = self.possible_values()

    def possible_values(self):
        a = "123456789"
        d, val = {}, {}
        for i in range(self.row):
            for j in range(self.col):
                ele = self.board[i][j]
                if ele != ".":
                    d[("r", i)] = d.get(("r", i), []) + [ele]
                    d[("c", j)] = d.get(("c", j), []) + [ele]
                    d[(i//3, j//3)] = d.get((i//3, j//3), []) + [ele]
                else:
                    val[(i,j)] = []
        for (i,j) in val.keys():
            inval = d.get(("r",i),[])+d.get(("c",j),[])+d.get((i/3,j/3),[])
            val[(i,j)] = [n for n in a if n not in inval ]
        return val

    def solve(self):
        if len(self.val)==0:
            return True
        kee = min(self.val.keys(), key=lambda x: len(self.val[x]))
        nums = self.val[kee]
        for n in nums:
            update = {kee:self.val[kee]}
            if self.valid_one(n, kee, update): # valid choice
                if self.solve(): # keep solving
                    return True
            self.undo(kee, update) # invalid choice or didn't solve it => undo
        return False

    def valid_one(self, n, kee, update):
        self.board[kee[0]][kee[1]] = n
        del self.val[kee]
        i, j = kee
        for ind in self.val.keys():
            if n in self.val[ind]:
                if ind[0]==i or ind[1]==j or (ind[0]/3,ind[1]/3)==(i/3,j/3):
                    update[ind] = n
                    self.val[ind].remove(n)
                    if len(self.val[ind])==0:
                        return False
        return True

    def undo(self, kee, update):
        self.board[kee[0]][kee[1]]="."
        for k in update:
            if k not in self.val:
                self.val[k]= update[k]
            else:
                self.val[k].append(update[k])
        return None

    def __str__(self):
        """[summary]
        Generates a board representation as string.

        Returns:
            [str] -- [board representation]
        """

        resp = ""
        for i in range(self.row):
            for j in range(self.col):
                resp += " {0} ".format(self.board[i][j])
            resp += "\n"
        return resp



"""
You are given a m x n 2D grid initialized with these three possible values:
    -1: A wall or an obstacle.
    0: A gate.
    INF: Infinity means an empty room. We use the value 2^31 - 1 = 2147483647 to represent INF 
         as you may assume that the distance to a gate is less than 2147483647.
Fill the empty room with distance to its nearest gate.
If it is impossible to reach a gate, it should be filled with INF.

For example, given the 2D grid:
    INF -1  0   INF
    INF INF INF -1
    INF -1  INF -1
    0   -1  INF INF
After running your function, the 2D grid should be:
    3   -1  0   1
    2   2   1   -1
    1   -1  2   -1
    0   -1  3   4
"""

def walls_and_gates(rooms):
    for i in range(len(rooms)):
        for j in range(len(rooms[0])):
            if rooms[i][j] == 0:
                dfs(rooms, i, j, 0)


def dfs(rooms, i, j, depth):
    if (i < 0 or i >= len(rooms)) or (j < 0 or j >= len(rooms[0])):
        return  # out of bounds
    if rooms[i][j] < depth:
        return  # crossed
    rooms[i][j] = depth
    dfs(rooms, i+1, j, depth+1)
    dfs(rooms, i-1, j, depth+1)
    dfs(rooms, i, j+1, depth+1)
    dfs(rooms, i, j-1, depth+1)



"""
Computing binomial coefficients is non optimization problem but can be solved using dynamic programming. 

Binomial coefficients are represented by C(n, k) or (nk) and can be used to represent the coefficients of a binomail:
```
(a + b)n  = C(n, 0)an + ... + C(n, k)an-kbk + ... + C(n, n)bn
```

The recursive relation is defined by the prior power
```
C(n, k) = C(n-1, k-1) + C(n-1, k) for n > k > 0

IC C(n, 0) = C(n, n) = 1
```

### Algorithm Binomial(n, k)

```
for i ← 0 to n do  // fill out the table row wise

for i = 0 to min(i, k) do

if j==0 or j==i then C[i, j] ← 1  // IC

else C[i, j] ← C[i-1, j-1] + C[i-1, j]  // recursive relation

return C[n, k]
```

The cost of the algorithm is filing out the table. Addition is the basic operation. Because k ≤ n, the sum needs to be split into two parts because only the half the table needs to be filled out for i < k and remaining part of the table is filled out across the entire row.


A(n, k) = sum for upper triangle + sum for the lower rectangle

```
= ∑i=1k ∑j=1i-1 1 + ∑i=1n ∑j=1k 1

= ∑i=1k (i-1) + ∑i=1n k

= (k-1)k/2 + k(n-k) ε Θ(nk)
```
"""
def bino(n,k):
    C = [[0 for i in range(k+1)] for i in range(n+1)]

    for i in range(n+1):
        for j in range(min(i,k)+1):
            if j in (0,i):
                C[i][j] = 1
            else:
                C[i][j] = C[i-1][j] + C[i-1][j-1]
    return C[n][k]


'''
Use dynamic programming (DP) to solve 0/1 knapsack problem
Time complexity: O(nW), where n is number of items and W is capacity

```
knapsack_dp(values,weights,n_items,capacity,return_all=False)

Input arguments:

  1. values: a list of numbers in either int or float, specifying the values of items
  2. weights: a list of int numbers specifying weights of items
  3. n_items: an int number indicating number of items
  4. capacity: an int number indicating the knapsack capacity
  5. return_all: whether return all info, defaulty is False (optional)

Return:

  1. picks: a list of numbers storing the positions of selected items
  2. max_val: maximum value (optional)
```
'''

import numpy as np

def knapsack_dp(values,weights,n_items,capacity,return_all=False):
    check_inputs(values,weights,n_items,capacity)

    table = np.zeros((n_items+1,capacity+1),dtype=np.float32)
    keep = np.zeros((n_items+1,capacity+1),dtype=np.float32)

    for i in range(1,n_items+1):
        for w in range(0,capacity+1):
            wi = weights[i-1] # weight of current item
            vi = values[i-1] # value of current item
            if (wi <= w) and (vi + table[i-1,w-wi] > table[i-1,w]):
                table[i,w] = vi + table[i-1,w-wi]
                keep[i,w] = 1
            else:
                table[i,w] = table[i-1,w]

    picks = []
    K = capacity

    for i in range(n_items,0,-1):
        if keep[i,K] == 1:
            picks.append(i)
            K -= weights[i-1]

    picks.sort()
    picks = [x-1 for x in picks] # change to 0-index

    if return_all:
        max_val = table[n_items,capacity]
        return picks,max_val
    return picks

def check_inputs(values,weights,n_items,capacity):
    # check variable type
    assert(isinstance(values,list))
    assert(isinstance(weights,list))
    assert(isinstance(n_items,int))
    assert(isinstance(capacity,int))
    # check value type
    assert(all(isinstance(val,int) or isinstance(val,float) for val in values))
    assert(all(isinstance(val,int) for val in weights))
    # check validity of value
    assert(all(val >= 0 for val in weights))
    assert(n_items > 0)
    assert(capacity > 0)


"""
Selection sort is a simple sorting algorithm. This sorting algorithm is an in-place 
comparison-based algorithm in which the list is divided into two parts, the sorted 
part at the left end and the unsorted part at the right end. Initially, the sorted 
part is empty and the unsorted part is the entire list.

The smallest element is selected from the unsorted array and swapped with the 
leftmost element, and that element becomes a part of the sorted array. This process 
continues moving unsorted array boundary by one element to the right.

This algorithm is not suitable for large data sets as its average and worst case 
complexities are of Ο(n2), where n is the number of items.

### Algorithm

```
Step 1 − Set MIN to location 0
Step 2 − Search the minimum element in the list
Step 3 − Swap with value at location MIN
Step 4 − Increment MIN to point to next element
Step 5 − Repeat until list is sorted
```

### Pseudocode

```python
procedure selection sort 
   list  : array of items
   n     : size of list

   for i = 1 to n - 1
   /* set current element as minimum*/
      min = i    
  
      /* check the element to be minimum */

      for j = i+1 to n 
         if list[j] < list[min] then
            min = j;
         end if
      end for

      /* swap the minimum element with the current element*/
      if indexMin != i  then
         swap list[min] and list[i]
      end if
   end for
	
end procedure
```
"""
def sort(array):
    for i in range(0, len(array) - 1):
        smallest = i
        for j in range(i + 1, len(array)):
            if array[j] < array[smallest]:
                smallest = j
        array[i], array[smallest] = array[smallest], array[i]
    return array


'''
The naïve approach tests all the possible placement of Pattern P [1.......m] relative 
to text T [1......n]. We try shift s = 0, 1.......n-m, successively and for each shifts. 
Compare T [s+1.......s+m] to P [1......m].

The naïve algorithm finds all valid shifts using a loop that checks the condition 
P[1.......m] = T [s+1.......s+m] for each of the n - m +1 possible value of s.

```
NAIVE-STRING-MATCHER (T, P)
 1. n ← length [T]
 2. m ← length [P]
 3. for s ← 0 to n -m
 4. do if P [1.....m] = T [s + 1....s + m]
 5. then print "Pattern occurs with shift" s
```

### Analysis: 

This for loop from 3 to 5 executes for n-m + 1(we need at least m characters at 
the end) times and in iteration we are doing m comparisons. So the total complexity is 
O (n-m+1).
'''
def match(text,pat):
    m = len(text)
    n = len(pat)
    pos = []
    for i in range(m-(n-1)):
        if text[i:i+n] == pat:
            pos.append(i)
    if len(pos):
        return pos
    else:
        return "No Match Found"



name="sorting"

from .bubble import sort
from .heap import sort
from .gnome import sort
from .counting import sort
from .comb import sort
from .cocktail import sort
from .bucket import sort
from .bogo import sort
from .shell import sort
from .selection import sort
from .radix import sort
from .quick import sort
from .merge import sort
from .insertion import sort


"""
Numbers can be regarded as product of its factors. For example,
8 = 2 x 2 x 2;
  = 2 x 4.


Write a function that takes an integer n and return all possible combinations
of its factors.Numbers can be regarded as product of its factors. For example,
8 = 2 x 2 x 2;
  = 2 x 4.

Examples:
input: 1
output:
[]


input: 37
output:
[]

input: 32
output:
[
  [2, 16],
  [2, 2, 8],
  [2, 2, 2, 4],
  [2, 2, 2, 2, 2],
"""
def get_factors(n):
    """[summary]
    
    Arguments:
        n {[int]} -- [to analysed number]
    
    Returns:
        [list of lists] -- [all factors of the number n]
    """

    def factor(n, i, combi, res):
        """[summary]
        helper function

        Arguments:
            n {[int]} -- [number]
            i {[int]} -- [to tested divisor]
            combi {[list]} -- [catch divisors]
            res {[list]} -- [all factors of the number n]
        
        Returns:
            [list] -- [res]
        """

        while i * i <= n:
            if n % i == 0:
                res += combi + [i, int(n/i)],
                factor(n/i, i, combi+[i], res)
            i += 1
        return res
    return factor(n, 2, [], [])


def get_factors_iterative1(n):
    """[summary]
    Computes all factors of n.
    Translated the function get_factors(...) in
    a call-stack modell.

    Arguments:
        n {[int]} -- [to analysed number]
    
    Returns:
        [list of lists] -- [all factors]
    """

    todo, res = [(n, 2, [])], []
    while todo:
        n, i, combi = todo.pop()
        while i * i <= n:
            if n % i == 0:
                res += combi + [i, n//i],
                todo.append((n//i, i, combi+[i])),
            i += 1
    return res


def get_factors_iterative2(n):
    """[summary]
    analog as above

    Arguments:
        n {[int]} -- [description]
    
    Returns:
        [list of lists] -- [all factors of n]
    """

    ans, stack, x = [], [], 2
    while True:
        if x > n // x:
            if not stack:
                return ans
            ans.append(stack + [n])
            x = stack.pop()
            n *= x
            x += 1
        elif n % x == 0:
            stack.append(x)
            n //= x
        else:
            x += 1



"""
Linear search is used on a collections of items. It relies on the technique of 
traversing a list from start to end by exploring properties of all the elements 
that are found on the way.

For example, consider an array of integers of size N. You should find and print 
the position of all the elements with value x. Here, the linear search is based 
on the idea of matching each element from the beginning of the list to the end 
of the list with the integer x, and then printing the position of the element 
if the condition is `True`.

### Psuedo Code:

The pseudo code for this example is as follows :

```
for(start to end of array)
{
    if (current_element equals to 5)  
    {
        print (current_index);
    }
}
```

### Time Complexity:

The time complexity of the linear search is 
O(N) because each element in an array is compared only once.

"""

def search(item,array):
	for i in range(len(list(array))):
		if array[i] == item:
			return i
	return "Not Found"


"""
### Longest Common Subsequence

If a set of sequences are given, the longest common subsequence problem is to find a common subsequence of all the sequences that is of maximal length.

The longest common subsequence problem is a classic computer science problem, the basis of data comparison programs such as the diff-utility, and has applications in bioinformatics. It is also widely used by revision control systems, such as SVN and Git, for reconciling multiple changes made to a revision-controlled collection of files.

### Naïve Method

Let X be a sequence of length m and Y a sequence of length n. Check for every subsequence of X whether it is a subsequence of Y, and return the longest common subsequence found.

There are 2m subsequences of X. Testing sequences whether or not it is a subsequence of Y takes O(n) time. Thus, the naïve algorithm would take O(n2m) time.

### Dynamic Programming

Let X = < x1, x2, x3,…, xm > and Y = < y1, y2, y3,…, yn > be the sequences. To compute the length of an element the following algorithm is used.

In this procedure, table C[m, n] is computed in row major order and another table B[m,n] is computed to construct optimal solution.

### Algorithm:

```
LCS-Length-Table-Formulation (X, Y)
m := length(X) 
n := length(Y) 
for i = 1 to m do 
   C[i, 0] := 0 
for j = 1 to n do 
   C[0, j] := 0 
for i = 1 to m do 
   for j = 1 to n do 
      if xi = yj 
         C[i, j] := C[i - 1, j - 1] + 1 
         B[i, j] := ‘D’ 
      else 
         if C[i -1, j] ≥ C[i, j -1] 
            C[i, j] := C[i - 1, j] + 1 
            B[i, j] := ‘U’ 
         else 
         C[i, j] := C[i, j - 1]
         B[i, j] := ‘L’ 
return C and B
```

### Analysis

To populate the table, the outer for loop iterates m times and the inner for loop iterates n times. Hence, the complexity of the algorithm is O(m, n), where m and n are the length of two strings.
"""
def lcs(x,y):
    m= len(x)
    n= len(y)
    # print(m,n)
    k = [[0 for i in range(n)] for i in range(m)]
    for i in range(m):
        for j in range(n):
            if i==0 or y==0:
                # print(i,j)
                k[i][j] = 0
            if x[i-1]==y[j-1]:
                k[i][j] = 1+k[i-1][j-1]
            else:
                k[i][j] = max(k[i-1][j], k[i][j-1])
    
    return k[m-1][n-1]


"""
BogoSort also known as permutation sort, stupid sort, slow sort, shotgun sort or 
monkey sort is a particularly ineffective algorithm based on generate and test 
paradigm. The algorithm successively generates permutations of its input until 
it finds one that is sorted.

Bogosort simply shuffles a collection randomly until it is sorted.

"Bogosort" is a perversely inefficient algorithm only used as an in-joke.

### Complexity:

Its average run-time is   O(n!)   because the chance that any given shuffle of a set will end up in sorted order is about one in   n   factorial,   and the worst case is infinite since there's no guarantee that a random shuffling will ever produce a sorted sequence.

Its best case is   O(n)   since a single pass through the elements may suffice to order them.


### Pseudocode:

```
while not InOrder(list) do
   Shuffle(list)
done
```
"""
import random
 
def bogosort(l):
    while not in_order(l):
        random.shuffle(l) # throw all elements
    return l
 
"""
For example, if bogosort is used to sort a deck of cards, it would consist of 
checking if the deck were in order, and if it were not, one would throw the 
deck into the air, pick the cards up at random, and repeat the process until 
the deck is sorted.
"""

def in_order(l):
    if not l:
        return True
    last = l[0]
    for x in l[1:]:
        if x < last:
            return False
        last = x
    return True


"""
### Complexity

- Worst case time complexity: O((N+1)!)
- Average case time complexity: O((N+1)!)
- Best case time complexity: O(N)
- Space complexity: O(1)

"""

def sort(array):
    return bogosort(array)



"""
Bubble Sort is a simple algorithm which is used to sort a given set of n elements 
provided in form of an array with n number of elements. Bubble Sort compares all 
the element one by one and sort them based on their values.

If the given array has to be sorted in ascending order, then bubble sort will start 
by comparing the first element of the array with the second element, if the first 
element is greater than the second element, it will swap both the elements, and then 
move on to compare the second and the third element, and so on.

If we have total n elements, then we need to repeat this process for n-1 times.

It is known as bubble sort, because with every complete iteration the largest element 
in the given array, bubbles up towards the last place or the highest index, just like 
a water bubble rises up to the water surface.

Sorting takes place by stepping through all the elements one-by-one and comparing it 
with the adjacent element and swapping them if required.

### Implementing Bubble Sort Algorithm

Following are the steps involved in bubble sort(for sorting a given array in ascending order):

- Starting with the first element(index = 0), compare the current element with the next element of the array.
- If the current element is greater than the next element of the array, swap them.
- If the current element is less than the next element, move to the next element. Repeat Step 1.

### Following are the Time and Space complexity for the Bubble Sort algorithm.

- Worst Case Time Complexity [ Big-O ]: O(n2)
- Best Case Time Complexity [Big-omega]: O(n)
- Average Time Complexity [Big-theta]: O(n2)
- Space Complexity: O(1)

"""
def sort(array):
    for i in range(len(array)):
        for j in range(len(array)-1,i,-1):
            if array[j] < array[j-1]:
                array[j-1], array[j] = array[j], array[j-1]
        
    return array


"""
Radix sort is a small method that many people intuitively use when alphabetizing a large list 
of names. Specifically, the list of names is first sorted according to the first letter of 
each name, that is, the names are arranged in 26 classes.

Intuitively, one might want to sort numbers on their most significant digit. However, Radix 
sort works counter-intuitively by sorting on the least significant digits first. On the first 
pass, all the numbers are sorted on the least significant digit and combined in an array. 
Then on the second pass, the entire numbers are sorted again on the second least significant 
digits and combined in an array and so on.

### Algorithm: 

```
Radix-Sort (list, n) 
shift = 1 
for loop = 1 to keysize do 
   for entry = 1 to n do 
      bucketnumber = (list[entry].key / shift) mod 10 
      append (bucket[bucketnumber], list[entry]) 
   list = combinebuckets() 
   shift = shift * 10 
```

### Analysis
Each key is looked at once for each digit (or letter if the keys are alphabetic) of the longest 
key. Hence, if the longest key has m digits and there are n keys, radix sort has order O(m.n).

However, if we look at these two values, the size of the keys will be relatively small when 
compared to the number of keys. For example, if we have six-digit keys, we could have a 
million different records.

Here, we see that the size of the keys is not significant, and this algorithm is of linear 
complexity O(n).
"""
def radix_sort(array, base=10):
    if array == []:
        return
 
    def key_factory(digit, base):
        def key(array, index):
            return ((array[index]//(base**digit)) % base)
        return key
    largest = max(array)
    exp = 0
    while base**exp <= largest:
        array = counting_sort(array, base - 1, key_factory(exp, base))
        exp = exp + 1
    return array
 
def counting_sort(array, largest, key):
    c = [0]*(largest + 1)
    for i in range(len(array)):
        c[key(array, i)] = c[key(array, i)] + 1
 
    # Find the last index for each element
    c[0] = c[0] - 1 # to decrement each element for zero-based indexing
    for i in range(1, largest + 1):
        c[i] = c[i] + c[i - 1]
 
    result = [None]*len(array)
    for i in range(len(array) - 1, -1, -1):
        result[c[key(array, i)]] = array[i]
        c[key(array, i)] = c[key(array, i)] - 1
 
    return result

def sort(array):
    return radix_sort(array)


"""
Fibonacci series generates the subsequent number by adding two previous numbers. Fibonacci series starts from two numbers − F0 & F1. The initial values of F0 & F1 can be taken 0, 1 or 1, 1 respectively.

Fibonacci series satisfies the following conditions −
```
Fn = Fn-1 + Fn-2
```
Hence, a Fibonacci series can look like this −
```
F8 = 0 1 1 2 3 5 8 13
```
or, this −
```
F8 = 1 1 2 3 5 8 13 21
```
"""
# Fibonacci Series using Dynamic Programming 
def fibonacci(n): 
	
	# Taking 1st two fibonacci nubers as 0 and 1 
	FibArray = [0, 1] 
	
	while len(FibArray) < n + 1: 
		FibArray.append(0) 
	
	if n <= 1: 
		return n 
	else: 
		if FibArray[n - 1] == 0: 
			FibArray[n - 1] = fibonacci(n - 1) 

		if FibArray[n - 2] == 0: 
			FibArray[n - 2] = fibonacci(n - 2) 
			
	FibArray[n] = FibArray[n - 2] + FibArray[n - 1] 
	return FibArray[n] 
 



"""
Quick sort is a highly efficient sorting algorithm and is based on partitioning of array of data 
into smaller arrays. A large array is partitioned into two arrays one of which holds values 
smaller than the specified value, say pivot, based on which the partition is made and another 
array holds values greater than the pivot value.

Quick sort partitions an array and then calls itself recursively twice to sort the two resulting 
subarrays. This algorithm is quite efficient for large-sized data sets as its average and worst 
case complexity are of Ο(n2), where n is the number of items.

### Quick Sort Algorithm

Using pivot algorithm recursively, we end up with smaller possible partitions. Each partition is 
then processed for quick sort. We define recursive algorithm for quicksort as follows −

```
Step 1 − Make the right-most index value pivot
Step 2 − partition the array using pivot value
Step 3 − quicksort left partition recursively
Step 4 − quicksort right partition recursively
```


### Quick Sort Pivot Algorithm

Based on our understanding of partitioning in quick sort, we will now try to write an algorithm for 
it, which is as follows.

```
Step 1 − Choose the highest index value has pivot

Step 2 − Take two variables to point left and right 
         of the list excluding pivot

Step 3 − left points to the low index

Step 4 − right points to the high

Step 5 − while value at left is less than pivot move 
         right

Step 6 − while value at right is greater than pivot
         move left

Step 7 − if both step 5 and step 6 does not match 
         swap left and right

Step 8 − if left ≥ right, the point where they met 
         is new pivot
```

"""
def quicksort(array, start, end):
    '''Sorts the list from indexes start to end - 1 inclusive.'''
    if end - start > 1:
        p = partition(array, start, end)
        quicksort(array, start, p)
        quicksort(array, p + 1, end)
    return array 

def partition(array, start, end):
    pivot = array[start]
    i = start + 1
    j = end - 1
 
    while True:
        while (i <= j and array[i] <= pivot):
            i = i + 1
        while (i <= j and array[j] >= pivot):
            j = j - 1
 
        if i <= j:
            array[i], array[j] = array[j], array[i]
        else:
            array[start], array[j] = array[j], array[start]
            return j

def sort(array):
    return quicksort(array, 0, len(array))


"""
The basic idea of comb sort and the bubble sort is same. In other words, 
comb sort is an improvement on the bubble sort. In the bubble sorting 
technique, the items are compared with the next item in each phase. But 
for the comb sort, the items are sorted in a specific gap. After completing 
each phase, the gap is decreased. The decreasing factor or the shrink factor 
for this sort is 1.3. It means that after completing each phase the gap is 
divided by 1.3.

### The complexity of Comb Sort Technique

- Time Complexity: O(n log n) for the best case. O(n^2/2^p) (p is a number of increment) 
for average case and O(n^2) for the worst case.

- Space Complexity: O(1)

### Input and Output

```
Input:
A list of unsorted data: 108 96 23 74 12 56 85 42 13 47

Output:
Array before Sorting: 108 96 23 74 12 56 85 42 13 47
Array after Sorting: 12 13 23 42 47 56 74 85 96 108
```

### Algorithm

```
CombSort(array, size)
Input: An array of data, and the total number in the array

Output: The sorted Array

Begin
   gap := size
   flag := true
   while the gap ≠ 1 OR flag = true do
      gap = floor(gap/1.3) //the the floor value after division
      if gap < 1 then
         gap := 1
      flag = false

      for i := 0 to size – gap -1 do
         if array[i] > array[i+gap] then
            swap array[i] with array[i+gap]
            flag = true;
      done
   done
End
```
"""
def comb_sort(array):
    def swap(i, j):
        array[i], array[j] = array[j], array[i]
 
    gap = len(array)
    shrink = 1.3
 
    no_swap = False
    while not no_swap:
        gap = int(gap/shrink)
 
        if gap < 1:
            gap = 1
            no_swap = True
        else:
            no_swap = False
 
        i = 0
        while i + gap < len(array):
            if array[i] > array[i + gap]:
                swap(i, i + gap)
                no_swap = False
            i = i + 1
    return array

def sort(array):
    return comb_sort(array)


"""
Given a 2d grid map of '1's (land) and '0's (water),
count the number of islands.
An island is surrounded by water and is formed by
connecting adjacent lands horizontally or vertically.
You may assume all four edges of the grid are all surrounded by water.

Example 1:

11110
11010
11000
00000
Answer: 1

Example 2:

11000
11000
00100
00011
Answer: 3
"""

def num_islands(grid):
    count = 0
    for i in range(len(grid)):
        for j, col in enumerate(grid[i]):
            if col == 1:
                dfs(grid, i, j)
                count += 1
    return count


def dfs(grid, i, j):
    if (i < 0 or i >= len(grid)) or (j < 0 or j >= len(grid[0])):
        return
    if grid[i][j] != 1:
        return
    grid[i][j] = 0
    dfs(grid, i+1, j)
    dfs(grid, i-1, j)
    dfs(grid, i, j+1)
    dfs(grid, i, j-1)



"""
Shell sort is a highly efficient sorting algorithm and is based on insertion sort algorithm. 
This algorithm avoids large shifts as in case of insertion sort, if the smaller value is to 
the far right and has to be moved to the far left.

This algorithm uses insertion sort on a widely spread elements, first to sort them and then 
sorts the less widely spaced elements. This spacing is termed as interval. This interval is 
calculated based on Knuth's formula as −

Knuth's Formula
h = h * 3 + 1
where −
   h is interval with initial value 1

This algorithm is quite efficient for medium-sized data sets as its average and worst-case 
complexity of this algorithm depends on the gap sequence the best known is Ο(n), where n is 
the number of items. And the worst case space complexity is O(n).

### Algorithm

Following is the algorithm for shell sort.

```
Step 1 − Initialize the value of h
Step 2 − Divide the list into smaller sub-list of equal interval h
Step 3 − Sort these sub-lists using insertion sort
Step 3 − Repeat until complete list is sorted
```

### Pseudocode

Following is the pseudocode for shell sort.

```python
procedure shellSort()
   A : array of items 
	
   /* calculate interval*/
   while interval < A.length /3 do:
      interval = interval * 3 + 1	    
   end while
   
   while interval > 0 do:

      for outer = interval; outer < A.length; outer ++ do:

      /* select value to be inserted */
      valueToInsert = A[outer]
      inner = outer;

         /*shift element towards right*/
         while inner > interval -1 && A[inner - interval] >= valueToInsert do:
            A[inner] = A[inner - interval]
            inner = inner - interval
         end while

      /* insert the number at hole position */
      A[inner] = valueToInsert

      end for

   /* calculate interval*/
   interval = (interval -1) /3;	  

   end while
   
end procedure
```
"""
def gaps(size):
    # uses the gap sequence 2^k - 1: 1, 3, 7, 15, 31, ...
    length = size.bit_length()
    for k in range(length - 1, 0, -1):
        yield 2**k - 1
 
 
def shell_sort(array):
    def insertion_sort_with_gap(gap):
        for i in range(gap, len(array)):
            temp = array[i]
            j = i - gap
            while (j >= 0 and temp < array[j]):
                array[j + gap] = array[j]
                j = j - gap
            array[j + gap] = temp
 
    for g in gaps(len(array)):
        insertion_sort_with_gap(g)
    
    return array

def sort(array):
    return shell_sort(array)


"""Like linear search and binary search, ternary search is a searching
technique that is used to determine the position of a specific value in an
array. In binary search, the sorted array is divided into two parts while in
ternary search, it is divided into 3 parts and then you determine in which part
the element exists.

Ternary search, like binary search, is a divide-and-conquer algorithm. It is mandatory
for the array (in which you will search for an element) to be sorted before you begin
the search. In this search, after each iteration it neglects ⅓ part of the array and
repeats the same operations on the remaining ⅔.

### Psuedo Code

```
int ternary_search(int l,int r, int x)
{
    if(r>=l)
    {
        int mid1 = l + (r-l)/3;
        int mid2 = r -  (r-l)/3;
        if(ar[mid1] == x)
            return mid1;
        if(ar[mid2] == x)
            return mid2;
        if(x<ar[mid1])
            return ternary_search(l,mid1-1,x);
        else if(x>ar[mid2])
            return ternary_search(mid2+1,r,x);
        else
            return ternary_search(mid1+1,mid2-1,x);

    }
    return -1;
}
```

### Complexity

O(log3N) , where N is the size of the array
"""


def ternary_search(left_index, right_index, search_key, array):
    if left_index <= right_index:
        mid_index1 = left_index + (right_index - left_index)//3
        mid_index2 = right_index - (right_index - left_index)//3
        # print(mid_index2, mid_index1, left_index, right_index)
        if array[mid_index1] == search_key:
            return mid_index1
        if array[mid_index2] == search_key:
            return mid_index2
        if search_key < array[mid_index1]:
            return ternary_search(left_index, mid_index1-1, search_key, array)
        elif search_key > array[mid_index2]:
            return ternary_search(mid_index2+1, right_index, search_key, array)
        else:
            return ternary_search(mid_index1+1, mid_index2-1, search_key, array)
    return "Not Found"

def search(item,array):
    left_index = 0
    right_index = len(array) - 1
    return ternary_search(left_index, right_index, item, array)



"""Knuth-Morris and Pratt introduce a linear time algorithm for the string
matching problem. A matching time of O (n) is achieved by avoiding comparison
with an element of 'S' that have previously been involved in comparison with
some element of the pattern 'p' to be matched. i.e., backtracking on the string
'S' never occurs.

### Components of KMP Algorithm:

#### 1. The Prefix Function (Π): 

The Prefix Function, Π for a pattern encapsulates knowledge
about how the pattern matches against the shift of itself. This information can be used
to avoid a useless shift of the pattern 'p.' In other words, this enables avoiding
backtracking of the string 'S.'

#### 2. The KMP Matcher:

With string 'S,' pattern 'p' and prefix function 'Π' as inputs, find
the occurrence of 'p' in 'S' and returns the number of shifts of 'p' after which occurrences
are found.

### The Prefix Function (Π)

Following pseudo code compute the prefix function, Π:

```
COMPUTE- PREFIX- FUNCTION (P)
 1. m ←length [P]               //'p' pattern to be matched
 2. Π [1] ← 0
 3. k ← 0
 4. for q ← 2 to m
 5. do while k > 0 and P [k + 1] ≠ P [q]
 6. do k ← Π [k]
 7. If P [k + 1] = P [q]
 8. then k← k + 1
 9. Π [q] ← k
 10. Return Π
```

### Running Time Analysis:

In the above pseudo code for calculating the prefix function, the for loop from step 4 to step
10  runs 'm' times. Step1 to Step3 take constant time. Hence the running time of computing
prefix function is O (m).

### The KMP Matcher:

The KMP Matcher with the pattern 'p,' the string 'S' and prefix function 'Π' as input, finds a
match of p in S. Following pseudo code compute the matching component of KMP algorithm:

```
KMP-MATCHER (T, P)
 1. n ← length [T]
 2. m ← length [P]
 3. Π← COMPUTE-PREFIX-FUNCTION (P)
 4. q ← 0               // numbers of characters matched
 5. for i ← 1 to n      // scan S from left to right
 6. do while q > 0 and P [q + 1] ≠ T [i]
 7. do q ← Π [q]                // next character does not match
 8. If P [q + 1] = T [i]
 9. then q ← q + 1              // next character matches
 10. If q = m                              // is all of p matched?
 11. then print "Pattern occurs with shift" i - m
 12. q ← Π [q]                          // look for the next match
```

### Running Time Analysis:

The for loop beginning in step 5 runs 'n' times, i.e., as long as the length of the string 'S.'
Since step 1 to step 4 take constant times, the running time is dominated by this for the loop.
Thus running time of the matching function is O (n).
"""

def match(txt, pat): 
    M = len(pat) 
    N = len(txt) 
    # create lps[] that will hold the longest prefix suffix 
    # values for pattern 
    lps = [0]*M 
    j = 0 # index for pat[] 
    #Preprocess the pattern (calculate lps[] array) 
    computeLPSArray(pat, M, lps) 
    i = 0 # index for txt[] 
    pos = []
    while i < N: 
        if pat[j] == txt[i]: 
            i += 1
            j += 1
        if j == M: 
            pos.append(i-j)
            j = lps[j-1] 
        # mismatch after j matches 
        elif i < N and pat[j] != txt[i]: 
            # Do not match lps[0..lps[j-1]] characters, 
            # they will match anyway 
            if j != 0: 
                j = lps[j-1] 
            else: 
                i += 1
    if len(pos):
        return pos
    else:
        return "No Match Found"

def computeLPSArray(pat, M, lps): 
	len = 0 # length of the previous longest prefix suffix 
	lps[0] # lps[0] is always 0 
	i = 1
	while i < M: 
		if pat[i]== pat[len]: 
			len += 1
			lps[i] = len
			i += 1
		else: 
			if len != 0: 
				len = lps[len-1] 
			else: 
				lps[i] = 0
				i += 1


"""
The simplest sort algorithm is not Bubble Sort..., it is not Insertion Sort..., it's Gnome Sort!

Gnome Sort is based on the technique used by the standard Dutch Garden Gnome (Du.: tuinkabouter). 
Here is how a garden gnome sorts a line of flower pots. 

Basically, he looks at the flower pot 
next to him and the previous one; if they are in the right order he steps one pot forward, 
otherwise he swaps them and steps one pot backwards. Boundary conditions: if there is no 
previous pot, he steps forwards; if there is no pot next to him, he is done.

### Psuedocode
```
void gnomesort(int n, int ar[]) {
	int i = 0;
	
	while (i < n) {
		if (i == 0 || ar[i-1] <= ar[i]) i++;
		else {int tmp = ar[i]; ar[i] = ar[i-1]; ar[--i] = tmp;}
	}
}
```
"""
def gnome_sort(array):
    for pos in range(1, len(array)):
        while (pos != 0 and array[pos] < array[pos - 1]):
            array[pos], array[pos - 1] = array[pos - 1], array[pos]
            pos = pos - 1
    return array
    
def sort(array):
    return gnome_sort(array)


"""
This is an in-place comparison-based sorting algorithm. Here, a sub-list is 
maintained which is always sorted. For example, the lower part of an array 
is maintained to be sorted. An element which is to be 'insert'ed in this 
sorted sub-list, has to find its appropriate place and then it has to be 
inserted there. Hence the name, insertion sort.

The array is searched sequentially and unsorted items are moved and inserted 
into the sorted sub-list (in the same array). This algorithm is not suitable 
for large data sets as its average and worst case complexity are of Ο(n2), 
where n is the number of items.

### Algorithm

Now we have a bigger picture of how this sorting technique works, so we can derive 
simple steps by which we can achieve insertion sort.

```
Step 1 − If it is the first element, it is already sorted. return 1;
Step 2 − Pick next element
Step 3 − Compare with all elements in the sorted sub-list
Step 4 − Shift all the elements in the sorted sub-list that is greater than the 
         value to be sorted
Step 5 − Insert the value
Step 6 − Repeat until list is sorted
```

#### Pseudocode

```python
procedure insertionSort( A : array of items )
   int holePosition
   int valueToInsert
	
   for i = 1 to length(A) inclusive do:
	
      /* select value to be inserted */
      valueToInsert = A[i]
      holePosition = i
      
      /*locate hole position for the element to be inserted */
		
      while holePosition > 0 and A[holePosition-1] > valueToInsert do:
         A[holePosition] = A[holePosition-1]
         holePosition = holePosition -1
      end while
		
      /* insert the number at hole position */
      A[holePosition] = valueToInsert
      
   end for
end procedure
```

"""
def sort(array):
    for i in range(1, len(array)):
        temp = array[i]
        j = i - 1
        while (j >= 0 and temp < array[j]):
            array[j + 1] = array[j]
            j = j - 1
        array[j + 1] = temp
    return array


"""
Cocktail sort is the variation of Bubble Sort which traverses the list in both 
directions alternatively. It is different from bubble sort in the sense that, 
bubble sort traverses the list in forward direction only, while this algorithm 
traverses in forward as well as backward direction in one iteration.

### Algorithm

In cocktail sort, one iteration consists of two stages:

- The first stage loop through the array like bubble sort from left to right. The adjacent elements are compared and if left element is greater than the right element, then we swap those elements. The largest element of the list is placed at the end of the array in the forward pass.
- The second stage loop through the array from the right most unsorted element to the left. The adjacent elements are compared and if right element is smaller than the left element then, we swap those elements. The smallest element of the list is placed at the beginning of the array in backward pass.

The process continues until the list becomes unsorted.

### Complexity

#### Best Case	
**Time Complexity**	        O(n2)

#### Average Case
**Time Complexity**	        O(n2)

#### Worst Case
**Time Complexity**	        O(n2)

**Space Complexity**	    O(1) 

"""
def cocktail_shaker_sort(array):
    def swap(i, j):
        array[i], array[j] = array[j], array[i]
 
    upper = len(array) - 1
    lower = 0
 
    no_swap = False
    while (not no_swap and upper - lower > 1):
        no_swap = True
        for j in range(lower, upper):
            if array[j + 1] < array[j]:
                swap(j + 1, j)
                no_swap = False
        upper = upper - 1
 
        for j in range(upper, lower, -1):
            if array[j - 1] > array[j]:
                swap(j - 1, j)
                no_swap = False
        lower = lower + 1
    return array

def sort(array):
    return cocktail_shaker_sort(array)


"""
Heaps can be used in sorting an array. In max-heaps, maximum element will always be 
at the root. Heap Sort uses this property of heap to sort the array.

Consider an array Arr which is to be sorted using Heap Sort.

- Initially build a max heap of elements in Arr.The root element, that is Arr[1], will 
contain maximum element of Arr. After that, swap this element with the last element of 
- Arr and heapify the max heap excluding the last element which is already in its correct 
position and then decrease the length of heap by one.
- Repeat the step 2, until all the elements are in their correct position.

### Psuedocode:

```
void heap_sort(int Arr[ ])

{
    int heap_size = N;

    build_maxheap(Arr);
    for(int i = N; i >= 2 ; i-- )
    {
        swap|(Arr[ 1 ], Arr[ i ]);
        heap_size = heap_size - 1;
        max_heapify(Arr, 1, heap_size);
    }
}
```

### Complexity:

`max_heapify` has complexity `O(logN)`, `build_maxheap` has complexity `O(N)` and we run `max_heapify` 
`N−1` times in `heap_sort` function, therefore complexity of `heap_sort` function is `O(NlogN)`.
"""
def heapsort(array):
    build_max_heap(array)
    for i in range(len(array) - 1, 0, -1):
        array[0], array[i] = array[i], array[0]
        max_heapify(array, index=0, size=i)
    return array
    
def parent(i):
    return (i - 1)//2
 
def left(i):
    return 2*i + 1
 
def right(i):
    return 2*i + 2
 
def build_max_heap(array):
    length = len(array)
    start = parent(length - 1)
    while start >= 0:
        max_heapify(array, index=start, size=length)
        start = start - 1
 
def max_heapify(array, index, size):
    l = left(index)
    r = right(index)
    if (l < size and array[l] > array[index]):
        largest = l
    else:
        largest = index
    if (r < size and array[r] > array[largest]):
        largest = r
    if (largest != index):
        array[largest], array[index] = array[index], array[largest]
        max_heapify(array, largest, size)

def sort(array):
    return heapsort(array)


'''
The Rabin-Karp string matching algorithm calculates a hash value for the pattern, 
as well as for each M-character subsequences of text to be compared. 

If the hash values are unequal, the algorithm will determine the hash value for next M-character 
sequence. If the hash values are equal, the algorithm will analyze the pattern and the
M-character sequence. 

In this way, there is only one comparison per text subsequence,
and character matching is only required when the hash values match.

```
RABIN-KARP-MATCHER (T, P, d, q)
 1. n ← length [T]
 2. m  ← length [P]
 3. h  ←  dm-1 mod q
 4. p ←  0
 5. t0 ←  0
 6. for i ← 1 to m
 7. do p ←  (dp + P[i]) mod q
 8. t0 ← (dt0+T [i]) mod q
 9. for s  ←  0 to n-m
 10. do if p = ts
 11. then if P [1.....m] = T [s+1.....s + m]
 12. then "Pattern occurs with shift" s
 13. If s < n-m
 14. then ts+1 ←  (d (ts-T [s+1]h)+T [s+m+1])mod q
```
### Complexity:

The running time of RABIN-KARP-MATCHER in the worst case scenario O ((n-m+1) m but it has a
good average case running time. If the expected number of strong shifts is small O (1) and 
prime q is chosen to be quite large, then the Rabin-Karp algorithm can be expected to run in 
time O (n+m) plus the time to require to process spurious hits.

'''
def match(text,pat):
    m = len(text)
    n = len(pat)

    p_hash = hash(pat)
    pos = []
    for i in range(m-(n-1)):
        if hash(text[i:i+n])==p_hash and text[i:i+n] == pat:
            pos.append(i)
    if len(pos):
        return pos
    else:
        return "No Match Found"


'''
Merge sort is a sorting technique based on divide and conquer technique. With worst-case time 
complexity being Ο(n log n), it is one of the most respected algorithms.

Merge sort first divides the array into equal halves and then combines them in a sorted manner.
### Algorithm

Merge sort keeps on dividing the list into equal halves until it can no more be divided. 
By definition, if it is only one element in the list, it is sorted. Then, merge sort 
combines the smaller sorted lists keeping the new list sorted too.

```
Step 1 − if it is only one element in the list it is already sorted, return.
Step 2 − divide the list recursively into two halves until it can no more be divided.
Step 3 − merge the smaller lists into new list in sorted order.
```

### Pseudocode

We shall now see the pseudocodes for merge sort functions. As our algorithms point out 
two main functions − divide & merge.

Merge sort works with recursion and we shall see our implementation in the same way.

```python
procedure mergesort( var a as array )
   if ( n == 1 ) return a

   var l1 as array = a[0] ... a[n/2]
   var l2 as array = a[n/2+1] ... a[n]

   l1 = mergesort( l1 )
   l2 = mergesort( l2 )

   return merge( l1, l2 )
end procedure

procedure merge( var a as array, var b as array )

   var c as array
   while ( a and b have elements )
      if ( a[0] > b[0] )
         add b[0] to the end of c
         remove b[0] from b
      else
         add a[0] to the end of c
         remove a[0] from a
      end if
   end while
   
   while ( a has elements )
      add a[0] to the end of c
      remove a[0] from a
   end while
   
   while ( b has elements )
      add b[0] to the end of c
      remove b[0] from b
   end while
   
   return c
	
end procedure
```
'''
def merge_sort(array, start, end):
    '''Sorts the list from indexes start to end - 1 inclusive.'''
    if end - start > 1:
        mid = (start + end)//2
        merge_sort(array, start, mid)
        merge_sort(array, mid, end)
        merge_list(array, start, mid, end)
        return array
        
def merge_list(array, start, mid, end):
    left = array[start:mid]
    right = array[mid:end]
    k = start
    i = 0
    j = 0
    while (start + i < mid and mid + j < end):
        if (left[i] <= right[j]):
            array[k] = left[i]
            i = i + 1
        else:
            array[k] = right[j]
            j = j + 1
        k = k + 1
    if start + i < mid:
        while k < end:
            array[k] = left[i]
            i = i + 1
            k = k + 1
    else:
        while k < end:
            array[k] = right[j]
            j = j + 1
            k = k + 1

def sort(array):
    return merge_sort(array, 0, len(array))


"""
### What is Bucket Sort ?

Bucket sort is a comparison sort algorithm that operates on elements by dividing 
them into different buckets and then sorting these buckets individually. Each 
bucket is sorted individually using a separate sorting algorithm or by applying 
the bucket sort algorithm recursively. Bucket sort is mainly useful when the 
input is uniformly distributed over a range.

#### Assume one has the following problem in front of them:

One has been given a large array of floating point integers lying uniformly 
between the lower and upper bound. This array now needs to be sorted. A simple 
way to solve this problem would be to use another sorting algorithm such as Merge 
sort, Heap Sort or Quick Sort. However, these algorithms guarantee a best case 
time complexity of O(NlogN). However, using bucket sort, the above task can be 
completed in O(N) time.

#### Psuedo Code 

```
void bucketSort(float[] a,int n)
{
    for(each floating integer 'x' in n)
    {
        insert x into bucket[n*x]; 
    }
    for(each bucket)
    {
        sort(bucket);
    }
}
```

### Time Complexity:

If one assumes that insertion in a bucket takes O(1) time, then steps 1 and 2 of the 
above algorithm clearly take O(n) time.

"""
def bucket_sort(array):
    largest = max(array)
    length = len(array)
    size = largest/length
 
    buckets = [[] for _ in range(length)]
    for i in range(length):
        j = int(array[i]/size)
        if j != length:
            buckets[j].append(array[i])
        else:
            buckets[length - 1].append(array[i])
 
    for i in range(length):
        insertion_sort(buckets[i])
 
    result = []
    for i in range(length):
        result = result + buckets[i]
 
    return result
 
def insertion_sort(array):
    for i in range(1, len(array)):
        temp = array[i]
        j = i - 1
        while (j >= 0 and temp < array[j]):
            array[j + 1] = array[j]
            j = j - 1
        array[j + 1] = temp

def sort(array):
    return bucket_sort(array)


"""
Counting sort is an efficient algorithm for sorting an array of elements that each have 
a nonnegative integer key, for example, an array, sometimes called a list, of positive 
integers could have keys that are just the value of the integer as the key, or a list 
of words could have keys assigned to them by some scheme mapping the alphabet to 
integers (to sort in alphabetical order, for instance). 

Unlike other sorting algorithms, 
such as mergesort, counting sort is an integer sorting algorithm, not a comparison based 
algorithm. 

While any comparison based sorting algorithm requires Omega(nlgn)Ω(nlgn) 
comparisons, counting sort has a running time of Theta(n)Θ(n) when the length of the 
input list is not much smaller than the largest key value, kk, in the list. Counting 
sort can be used as a subroutine for other, more powerful, sorting algorithms such as 
radix sort

### Complexity of Counting Sort

Counting sort has a O(k+n)O(k+n) running time.

The first loop goes through A, which has n elements. This step has a O(n)O(n) running 
time. The second loop iterates over k, so this step has a running time of O(k)O(k). The 
third loop iterates through AA, so again, this has a running time of O(n)O(n). Therefore, 
the counting sort algorithm has a running time of O(k+n)O(k+n).

Counting sort is efficient if the range of input data, kk, is not significantly greater 
than the number of objects to be sorted, nn.

Counting sort is a stable sort with a space complexity of **O(k + n)O(k+n).**
"""
def counting_sort(array, largest):
    c = [0]*(largest + 1)
    for i in range(len(array)):
        c[array[i]] = c[array[i]] + 1
 
    # Find the last index for each element
    c[0] = c[0] - 1 # to decrement each element for zero-based indexing
    for i in range(1, largest + 1):
        c[i] = c[i] + c[i - 1]
 
    result = [None]*len(array)
 
    # Though it is not required here,
    # it becomes necessary to reverse the list
    # when this function needs to be a stable sort
    for x in reversed(array):
        result[c[x]] = x
        c[x] = c[x] - 1
 
    return result

def sort(array):
    return counting_sort(array, max(array))


"""
Binary search is the most popular Search algorithm. It is efficient and also 
one of the most commonly used techniques that is used to solve problems.

If all the names in the world are written down together in order and you want 
to search for the position of a specific name, binary search will accomplish 
this in a maximum of 35 iterations.

Binary search works only on a sorted set of elements. To use binary search on 
a collection, the collection must first be sorted.

When binary search is used to perform operations on a sorted set, the number 
of iterations can always be reduced on the basis of the value that is being 
searched.

### Time complexity

As we dispose off one part of the search case during every step of binary search, 
and perform the search operation on the other half, this results in a worst case 
time complexity of O(log2N).

"""
def search(item,array):
    beg = 0
    end = len(array)
    while(end >= beg):
        mid = beg + ((end - beg)//2)
        # if mid goes out of bound report element not present
        if mid >= len(array):
            break
        if array[mid] > item:
            end = mid - 1
        elif array[mid] < item:
            beg = mid + 1
        else:
            return mid
    return "Not Found"